// ======================================================================
//
// Appear.cpp
// jeff grills
//
// copyright 1996 origin systems inc
//
// ======================================================================

#include "First.h"
#include "Appear.h"

#include "AnimAppear.h"
#include "Camera.h"
#include "DetailAppear.h"
#include "Extents.h"
#include "Game.h"
#include "HardPoints.h"
#include "Iff.h"
#include "MeshAppear.h"
#include "StepAppear.h"

#if USE_SIMPLE_APPEAR_NAMES
#include <stdio.h>
#endif

// ======================================================================

Appear *Appear::first;

// ======================================================================

inline bool Appear::found(const Appear *a, const CrcString *crcname, real scale)
{ //lint !e578
	return (a->scale == scale) && a->name && crcname->compareCrcString(*(a->crcName)) == 0; //lint !e777 // testing floats for equality
}

// ----------------------------------------------------------------------

Appear *Appear::find(const char *name, real scale, bool shouldAddUser)
{ //lint !e578
	Appear *appr;

	if (name)
	{
#if USE_SIMPLE_APPEAR_NAMES
		// check to see if the name already looks like it has a path and extension
		char buffer[128];

		if (!strchr(name, '\\') && !strchr(name, '.'))
		{
			sprintf(buffer, "mesh\\%s.iff", name);
			name = buffer;
		}
#endif

		CrcString crcname(name);

		// scan a list of loaded appears
		// -qq- maybe do some cksum stuff for a faster search
		for (appr = first; appr && !found(appr, &crcname, scale); appr = appr->nextAppear)
			;

		// if the appearance was not found, create it
		if (!appr)
		{
			Iff iff(name);
			appr = Game::createAppear(name, iff, scale);
			iff.close();
		
			// add this mesh to the loaded mesh list
			appr->nextAppear = first;
			first = appr;
		}
	}
	else
		appr = NULL;

	if (shouldAddUser)
		addUser(appr);

	return appr;
}

// ----------------------------------------------------------------------

Appear *Appear::addUser(Appear *appr)
{
	if (appr)
		appr->users++;
	return appr;
}

// ----------------------------------------------------------------------

void Appear::removeUser(Appear *appr)
{
	// check if nobody is using this mesh
	if (appr && --appr->users <= 0)
	{
		Appear *front, *back;

		// remove mesh from the loaded mesh list
		for (back = NULL, front = first; front && front != appr; back = front, front = front->nextAppear)
			;

		// if the appearance was found, remove it from the linked list
		if (front)
			if (back)
				back->nextAppear = front->nextAppear;
			else
				first = front->nextAppear;

		// delete it no matter what
		delete appr;
	}
}

// ----------------------------------------------------------------------

Appear *Appear::create(const char *name, Iff &iff, real scale)
{  //lint !e578
	Appear *appr;

	// load the correct appearance type
	iff.enterForm();
	switch (iff.parentName())
	{
		// this fixes the detail level madness.
		case TAG_0000:
		case TAG_0001:
		case TAG_0002:
		case TAG_0003:
		case TAG_0004:
		case TAG_0005:
		case TAG_0006:
		case TAG_0007:
		case TAG_0008:
		case TAG_0009:
			appr = Game::createAppear(name, iff, scale);
			break;

		case TAG_EMPT:
			appr = NULL;
			break;

		case TAG_MESH:
			appr = NEW MeshAppear(name, iff, scale);
			break;
					
		case TAG_DETA:
			appr = NEW DetailAppear(name, iff, scale);
			break;

		case TAG_ANIM:
			appr = NEW AnimAppear(name, iff, scale);
			break;

		case TAG_STEP:
			appr = NEW StepAppear(name, iff, scale);
			break;

		case TAG_NAME:
			{
				char buffer[128];

				iff.seekChunk(TAG_NAME);
					iff.read_string(buffer, sizeof(buffer));
				iff.exitChunk();

				appr = Appear::find(buffer, scale, false);
			}
			break;

		default:
			char tag[5];	
			appr = NULL;
			Iff::formatTag(iff.parentName(), tag);
			DEBUG_FATAL(true, ("Appear::create unknown appearance type in %s, tag %s (check mesh file)", name, tag));
	}
	iff.exitForm();

	return appr;
}

// ======================================================================

Appear::Appear(AppearId newId, const char *newName, real newScale)
{
	id             = newId;
	scale          = newScale;
	users          = 0;
	token          = reinterpret_cast<int>(this);
	needsRealAlter = false;

	extents        = NULL;
	hardPoints     = NULL;

	nextAppear     = NULL;

	if (newName)
	{
		name = MEM_OWN(strdup(newName));
		crcName = NEW CrcString(name);
	}
	else
	{
		name = NULL;
		crcName = NULL;
	}
}

// ----------------------------------------------------------------------

Appear::~Appear(void)
{
	delete crcName;

	delete [] name;
	name = NULL;

	Extents::removeUser(extents);
	extents = NULL;

	HardPoints::removeUser(hardPoints);
	hardPoints = NULL;

	nextAppear = NULL;
}

// ----------------------------------------------------------------------

void Appear::setHardPoints(HardPoints *newHardPoints)
{
	HardPoints::removeUser(hardPoints);
	hardPoints = newHardPoints;
	HardPoints::addUser(hardPoints);
}

// ----------------------------------------------------------------------

void Appear::setExtents(Extents *newExtents)
{
	Extents::removeUser(extents);
	extents = newExtents;
	Extents::addUser(extents);
}

// ----------------------------------------------------------------------

void Appear::loadShared(Iff &iff, const char *filename)
{
	if (iff.seekForm(TAG_HARD, Iff::FlagOptional))
	{
		setHardPoints(NEW HardPoints(iff, filename));
		iff.exitForm();
	}

	// load in the collision information
	if (iff.seekForm(TAG_COLL, Iff::FlagOptional))
	{
		setExtents(NEW Extents(iff));
		iff.exitForm();
	}
}

// ----------------------------------------------------------------------

void Appear::realAlter(real)
{
}

// ----------------------------------------------------------------------

const Appear *Appear::getRenderAppear(Camera &, const Object *) const
{
	return this;
}

// ----------------------------------------------------------------------

void Appear::render(Camera &, const Xform &) const
{
	if (DFlag(DF_E_WIRE_HARDPOINTS) && hardPoints)
		hardPoints->render();

#if 0
	if (DFlag(DF_E_WIRE_EXTENTS) && extents)
		extents->render();
#endif

}

// ======================================================================
